@@ -1,6 +1,6 @@
 #ifndef BOB_H_
 #define BOB_H_
 
-int bobMcBob();
+int bobMcBob(void);
 
 #endif
